/*
 * LM78 sensor driver for BeOS 
 *	 R4 port by Andrew Vessey (avessey@mindspring.com)
 *		- fixed lm78_read sprintf prob
 *		- moved code to R4 style bus manager read/writes
 *   original R3 port by Kazuki Sakamoto (sakamoto@cec.co.jp)
 */
 
/* -*- linux-c -*-
 *
 *  LM78 sensor driver for LINUX 
 *
 * Copyright (c) 1997 Alexander Larsson
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 * 
 * provides a /proc/sensors which shows
 * some information about your system.
 * Voltages, mb temperature and fan speeds
 * are currently reported.
 *
 * This should work on all ASUS TX97 cards
 * but it's only tested on a TX97-X.
 * It should also work on any card that has
 * a LM78 circuit, but some constants in
 * lm78.h probably has to be changed.
 *
 * please report any bugs/patches to me:
 *  alla@lysator.liu.se
 *
 * Reference:
 *   National Semiconductor
 *   LM78 Microprocessor System Hardware Monitor, September 1996.
 *   http://www.national.com/ds/LM/LM78.pdf
 *
 */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>
#include <stdio.h>

#define LM78_BASE		0x290
#define LM78_ADDRESS_REG	(LM78_BASE + 0x5)
#define LM78_DATA_REG		(LM78_BASE + 0x6)

#define	LM78_WAIT		1000	/* 1000us */

#define LM78_BUSY		0x80
#define LM78_TEST		0x20
#define	LM78_NOFAN		255

#define LM78_IN0 0x20     
#define LM78_IN1 0x21     
#define LM78_IN2 0x22     
#define LM78_IN3 0x23     
#define LM78_IN4 0x24     
#define LM78_MINUS_IN5 0x25
#define LM78_MINUS_IN6 0x26     
#define LM78_TEMPERATURE 0x27
#define LM78_FAN1 0x28
#define LM78_FAN2 0x29
#define LM78_FAN3 0x2A

#define LM78_FANDIV 0x47

/* Calculate rpms from tacho-count (at div=2) */
#define LM78_RPM(count) (337500/count)

#define	LM78_VCORE	LM78_IN0
#define	LM78_3_3V	LM78_IN2
#define	LM78_5V		LM78_IN3
#define	LM78_12V	LM78_IN4
#define	LM78_M12V	LM78_MINUS_IN5
#define	LM78_M5V	LM78_MINUS_IN6
		
#define	LM78_VCORE_A	200
#define	LM78_VCORE_B	125
#define	LM78_3_3V_A	200
#define	LM78_3_3V_B	125
#define	LM78_5V_A	334
#define	LM78_5V_B	125
#define	LM78_M5V_A	300
#define	LM78_M5V_B	125
#define	LM78_12V_A	758
#define	LM78_12V_B	125
#define	LM78_M12V_A	700
#define	LM78_M12V_B	125

/* -----
	null-terminated array of device names supported by this driver
----- */
static const char *name_list[] = {
	"misc/lm78",
	NULL
};

static status_t lm78_open(const char *, uint32, void **);
static status_t lm78_close(void *);
static status_t lm78_free(void *);
static status_t lm78_control(void *, uint32, void *, size_t);
static status_t lm78_read(void *, off_t, void *, size_t *);
static status_t lm78_write(void *, off_t, const void *, size_t *); 

int32 api_version = B_CUR_DRIVER_API_VERSION;
static long is_open;
isa_module_info * isa;

static inline void
LM78_WRITE(int addr, uint8 data)
{
	isa->write_io_8(LM78_ADDRESS_REG, addr);
	isa->write_io_8(LM78_DATA_REG, data);
}

static inline uint8
LM78_READ(int addr)
{
	isa->write_io_8(LM78_ADDRESS_REG, addr);
	return isa->read_io_8(LM78_DATA_REG);
}

static int
lm78_read_value(int addr) {
	while (isa->read_io_8(LM78_ADDRESS_REG) & LM78_BUSY)
		{
		spin(LM78_WAIT);
		}
	return LM78_READ(addr);
}

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
_EXPORT status_t
init_hardware(void)
{
	int ret;
	if (get_module(B_ISA_MODULE_NAME, (module_info **) &isa) <0)
		return ENOSYS;
	/* Verify that lm78 exists */
	isa->write_io_8(LM78_ADDRESS_REG, 0x20);
	ret = isa->read_io_8(LM78_ADDRESS_REG);
	isa->read_io_8(LM78_DATA_REG);
	if ((ret & 0x7f) != LM78_TEST)
	{	dprintf("NO lm78 found at %x\n",LM78_BASE);
		return  B_ERROR;
	}
		
	/* Set FAN1 and FAN2 to divide by 2 */
	LM78_WRITE(LM78_FANDIV, 0x50);
	put_module(B_ISA_MODULE_NAME);
	return  B_OK;
}

/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
_EXPORT status_t
init_driver(void)
{
	if (get_module(B_ISA_MODULE_NAME, (module_info **) &isa) <0)
		return ENOSYS;
	is_open = 0;
	return B_OK;
	
}

/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
_EXPORT void
uninit_driver(void)
{
	put_module(B_ISA_MODULE_NAME);
}

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */


static status_t
lm78_open(const char *name, uint32 flags, void **cookie)
{
	int id;
	id = lookup_device_name(name);
	if (id < 0){
		return EINVAL;}

	if (atomic_or(&is_open, 1) & 1)
	{
		return B_BUSY;}
	return B_OK;
}

/* ----------
	my_device_open - handle open() calls
----- */

/* ----------
	my_device_close - handle close() calls
----- */
static status_t
lm78_close(void *cookie)
{
	return B_OK;
}

/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
lm78_free(void *cookie)
{
	atomic_and(&is_open, ~1);
	return B_OK;
}


/* ----------
	my_device_control - handle ioctl calls
----- */
static status_t
lm78_control(void *cookie, uint32 op, void *data, size_t len)
{
	switch (op) {
	default:
		return EINVAL;
	}
	return B_OK;
}

/* ----------
	my_device_write - handle write() calls
----- */
status_t
lm78_write(void *cookie, off_t pos, const void *data, size_t *len)
{
	*len = 0;
	return B_IO_ERROR;
}


/* ----------

	lm78_read - handle read() calls
	This is where the biggest problem in the port from r3 to r4 appeared.
	Kept getting segment violations as it appears that sprintf does not return the 
	same value (number of chars written) as it did w/metrowerks compiler.
	Because of this had to add a lot of calls to strlen...
	
----- */
static status_t
lm78_read(void *cookie, off_t pos, void *data, size_t *len)
{
	
	int fan;
	int v;

	*len = 0;
    
	v = (lm78_read_value(LM78_VCORE) * LM78_VCORE_A) / LM78_VCORE_B;
	sprintf((char *)data + *len, "Vcore:  %d.%02dV\n", (v/100), (v%100) );
	*len = strlen((char *)data);

	v = (lm78_read_value(LM78_3_3V) * LM78_3_3V_A) / LM78_3_3V_B;
	sprintf((char *)data + *len, "+3.3V:  %d.%02dV\n", v/100, v%100 );
	*len = strlen((char *)data);

	v = (lm78_read_value(LM78_5V) * LM78_5V_A) / LM78_5V_B;	
	sprintf((char *)data + *len, "+5.0V:  %d.%02dV\n", v/100, v%100 );
	*len = strlen((char *)data);

	v = (lm78_read_value(LM78_M5V) * LM78_M5V_A) / LM78_3_3V_B;	
	sprintf((char *)data + *len, "-5.0V: -%d.%02dV\n", v/100, v%100 );
	*len = strlen((char *)data);

	v = (lm78_read_value(LM78_12V) * LM78_12V_A) / LM78_12V_B;	
	sprintf((char *)data + *len, "+12V :  %d.%02dV\n", v/100, v%100 );
	*len = strlen((char *)data);

	v = (lm78_read_value(LM78_M12V) * LM78_M12V_A) / LM78_M12V_B;	
	sprintf((char *)data + *len, "-12V : -%d.%02dV\n", v/100, v%100 );
	*len = strlen((char *)data);

	sprintf((char *)data + *len, "Mainboard temperature: %d degrees\n",
		lm78_read_value(LM78_TEMPERATURE));
	*len = strlen((char *)data);

	fan = LM78_READ(LM78_FAN1);
	sprintf((char *)data + *len, "Chassis fan: ");
	*len = strlen((char *)data);
	if (fan == LM78_NOFAN)
		sprintf((char *)data + *len, "not present\n");
	else
		sprintf((char *)data + *len, "%d rpm\n", LM78_RPM(fan));
	*len = strlen((char *)data);
	
	fan = LM78_READ(LM78_FAN2);
	sprintf((char *)data + *len, "CPU fan: ");
	*len = strlen((char *)data);
	if (fan == LM78_NOFAN)
		sprintf((char *)data + *len, "not present\n");
	else
		sprintf((char *)data + *len, "%d rpm\n", LM78_RPM(fan));
	*len = strlen((char *)data);
	
	fan = LM78_READ(LM78_FAN3);
	sprintf((char *)data + *len, "Power fan: ");
	*len = strlen((char *)data);
	
	if (fan == LM78_NOFAN)
		sprintf((char *)data + *len, "not present\n");
	else
		sprintf((char *)data + *len, "%d rpm\n", LM78_RPM(fan));
	*len = strlen((char *)data);
	return B_OK;
}

/* -----
	function pointers for the device hooks entry points
----- */
_EXPORT static device_hooks lm78_hooks = {
	lm78_open,
	lm78_close,
	lm78_free,
	lm78_control,
	lm78_read,
	lm78_write
};

_EXPORT const char**
publish_devices()
{
	return name_list;
}

int
lookup_device_name(const char *name)
{	
	int i;
	for (i = 0; name_list[i]; i++)
		if (!strcmp (name_list[i], name))
			{
			return i;}
	return -1;		/* Invalid device name */
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */
_EXPORT device_hooks *
find_device(const char *name)
{
	if (lookup_device_name(name) >= 0)
		{
		return &lm78_hooks;	/* Return hooks list */
		}
	return NULL;		/* Device not found */
}








	








